// this file is part of the Quirk browser
// copyright 2021 by zamfofex
// license: GPLv3 or later

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <termio.h>
#include <fcntl.h>

#include "console.h"

static void qr_write(int fd, char *text, size_t length);
static char qr_char(int fd);
static unsigned char qr_color(int r, int g, int b);
static QrDisplaySize qr_console_size(QrDisplay display);
static void qr_console_close(QrDisplay display);
static void qr_console_display(QrDisplay display, QrRectangle *rect, QrText *text);
static void qr_console_out(QrDisplay display);
static QrNumber qr_console_advance(QrDisplay display, char *string);
static void *qr_console_in(void *display);
static void qr_console_free(QrDisplay display);

void qr_write(fd, text, length)
	int fd;
	char *text;
	size_t length;
{
	while (length > 0)
	{
		ssize_t count = write(fd, text, length);
		if (count < 0) return;
		length -= count;
		text += count;
	}
}

char qr_char(fd)
	int fd;
{
	char ch;
	ssize_t count = read(fd, &ch, 1);
	if (count > 0) return ch;
	else return '\0';
}

// todo: use this
unsigned char qr_color(r, g, b)
	int r, g, b;
{
	static unsigned char snaps[] = {0x00, 0x5F, 0x87, 0xAF, 0xD7, 0xFF, 0xFF};
	
	for (unsigned int i = 0 ; ; i++)
		if (r * 2 < snaps[i] + snaps[i + 1]) { r = i; break; }
	for (unsigned int i = 0 ; ; i++)
		if (g * 2 < snaps[i] + snaps[i + 1]) { g = i; break; }
	for (unsigned int i = 0 ; ; i++)
		if (b * 2 < snaps[i] + snaps[i + 1]) { b = i; break; }
	
	return r*36 + g*6 + b + 16;
}

void qr_console_close(display)
	QrDisplay display;
{
	QrConsole console = display->data;
	qr_write(console->out, "\ec", 2);
	tcsetattr(console->out, TCSANOW, &console->termios);
}

QrDisplaySize qr_console_size(display)
	QrDisplay display;
{
	QrConsole console = display->data;
	
	struct winsize size;
	ioctl(console->in, TIOCGWINSZ, &size);
	
	unsigned int width = size.ws_col - 1;
	unsigned int height = size.ws_row - 1;
	
	QrDisplaySize result = {width * 8, height * 16};
	return result;
}

void qr_console_display(display, rect, text)
	QrDisplay display;
	QrRectangle *rect;
	QrText *text;
{
	QrConsole console = display->data;
	console->rectangles = rect;
	console->text = text;
	qr_console_out(display);
}

void qr_console_out(display)
	QrDisplay display;
{
	QrConsole console = display->data;
	
	QrRectangle *rect = console->rectangles;
	QrText *text = console->text;
	
	unsigned int width = rect[0]->width / 8;
	unsigned int height = rect[0]->height / 16;
	
	console->screen = realloc(console->screen, height * (width + 1));
	console->screen_length = height * (width + 1);
	char (*screen)[height][width + 1] = console->screen;
	
	for (unsigned int y = 0 ; y < height ; y++)
	for (unsigned int x = 0 ; x < width ; x++)
		(*screen)[y][x] = ' ';
	for (unsigned int y = 0 ; y < height ; y++)
		(*screen)[y][width] = '\n';
	
	while (*text)
	{
		char *string = text[0]->string;
		QrNumber x = text[0]->x / 8;
		QrNumber y = text[0]->y / 16 - console->scroll;
		
		while (*string)
		{
			if (x < width && y < height)
				(*screen)[y][x++] = *string;
			string++;
		}
		
		text++;
	}
	
	qr_write(console->out, "\ec\e[48;5;255m\e[38;5;236m", 24);
	qr_write(console->out, console->screen, console->screen_length);
}

QrNumber qr_console_advance(display, string)
	QrDisplay display;
	char *string;
{
	return strlen(string) * 8;
}

void *qr_console_in(data)
	void *data;
{
	QrDisplay display = data;
	QrConsole console = display->data;
	
	for (;;)
	{
		switch (qr_char(console->in))
		{
			case 'W':
			case 'w':
				if (console->scroll != 0)
					console->scroll--;
				qr_console_out(display);
				break;
			case 'S':
			case 's':
				console->scroll++;
				qr_console_out(display);
				break;
			case 'Q':
			case 'q':
				// todo
				qr_console_close(display);
				exit(0);
				continue;
		}
	}
}

QrDisplay qr_console(in, out)
	int in, out;
{
	QrConsole console = malloc(sizeof *console);
	console->in = in;
	console->out = out;
	console->scroll = 0;
	console->screen = NULL;
	
	tcgetattr(in, &console->termios);
	struct termios new = console->termios;
	new.c_lflag &= ~(ICANON | ECHO);
	tcsetattr(in, TCSANOW, &new);
	
	QrDisplay display = malloc(sizeof *display);
	display->data = console;
	display->display = &qr_console_display;
	display->advance = &qr_console_advance;
	display->size = &qr_console_size;
	display->free = &qr_console_free;
	
	display->events = qr_queue();
	
	pthread_t thread;
	pthread_create(&thread, NULL, &qr_console_in, display);
	
	return display;
}

void qr_console_free(display)
	QrDisplay display;
{
	free(display->data);
	free(display);
}
