// this file is part of the Quirk browser
// copyright 2021 by zamfofex
// license: GPLv3 or later

#include <arpa/inet.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <poll.h>
#include <time.h>

#include "builtin.h"
#include "dom.h"
#include "../util/syntax.h"

static JSValue qr_builtin_print(JSContext *ctx, JSValue this, int argc, JSValue *argv);
static JSValue qr_builtin_encode(JSContext *ctx, JSValue this, int argc, JSValue *argv);
static JSValue qr_builtin_decode(JSContext *ctx, JSValue this, int argc, JSValue *argv);
static JSValue qr_builtin_new_context(JSContext *ctx, JSValue this, int argc, JSValue *argv);
static JSValue qr_builtin_run_script(JSContext *ctx, JSValue this, int argc, JSValue *argv);
static JSValue qr_builtin_now(JSContext *ctx, JSValue this, int argc, JSValue *argv);
static JSValue qr_builtin_curl(JSContext *ctx, JSValue this, int argc, JSValue *argv);
static size_t qr_builtin_curl_write(char *buffer, size_t one, size_t size, void *data);
static size_t qr_builtin_curl_header(char *buffer, size_t one, size_t size, void *data);
static JSValue qr_builtin_url(JSContext *ctx, JSValue this, int argc, JSValue *argv);

typedef struct QrJSValue *QrJSValue;
struct QrJSValue
{
	JSContext *ctx;
	JSValue value;
};

JSValue qr_builtin_print(ctx, this, argc, argv)
	JSContext *ctx;
	JSValue this;
	int argc;
	JSValue *argv;
{
	/*
	for (unsigned int i = 0 ; i < argc ; i++)
	{
		char *string = JS_ToCString(ctx, argv[i]);
		qr_report(string);
		qr_report(" ");
		JS_FreeCString(ctx, string);
	}
	qr_report("\n");
	*/
	
	return JS_UNDEFINED;
}

JSValue qr_builtin_encode(ctx, this, argc, argv)
	JSContext *ctx;
	JSValue this;
	int argc;
	JSValue *argv;
{
	char *string = JS_ToCString(ctx, argv[0]);
	qr_assert(string != NULL);
	JSValue value = JS_NewArrayBufferCopy(ctx, string, strlen(string));
	JS_FreeCString(ctx, string);
	return value;
}

JSValue qr_builtin_decode(ctx, this, argc, argv)
	JSContext *ctx;
	JSValue this;
	int argc;
	JSValue *argv;
{
	size_t length;
	char *buffer = JS_GetArrayBuffer(ctx, &length, argv[0]);
	char *string = malloc(length);
	qr_assert(string != NULL);
	memcpy(string, buffer, length);
	JSValue value = JS_NewStringLen(ctx, string, length);
	free(string);
	return value;
}

JSValue qr_builtin_new_context(ctx, this, argc, argv)
	JSContext *ctx;
	JSValue this;
	int argc;
	JSValue *argv;
{
	JSContext *other = JS_NewContext(JS_GetRuntime(ctx));
	JSValue global = JS_GetGlobalObject(other);
	JS_SetOpaque(global, other);
	return global;
}

JSValue qr_builtin_run_script(ctx, this, argc, argv)
	JSContext *ctx;
	JSValue this;
	int argc;
	JSValue *argv;
{
	JSContext *other = JS_GetOpaque(argv[0], 1);
	char *js = JS_ToCString(ctx, argv[1]);
	
	JSValue value = JS_Eval(other, js, strlen(js), "[anonymous]", JS_EVAL_TYPE_GLOBAL);
	
	JS_FreeCString(ctx, js);
	
	return value;
}

JSValue qr_builtin_now(ctx, this, argc, argv)
	JSContext *ctx;
	JSValue this;
	int argc;
	JSValue *argv;
{
	struct timespec spec;
	qr_assert(clock_gettime(CLOCK_MONOTONIC, &spec) == 0);
	
	uint32_t ms = spec.tv_sec * 1000 + spec.tv_nsec / 1000000;
	return JS_NewUint32(ctx, ms);
}

size_t qr_builtin_curl_write(buffer, one, size, data)
	char *buffer;
	size_t one;
	size_t size;
	void *data;
{
	QrJSValue callback = data;
	JSValue value = JS_NewArrayBufferCopy(callback->ctx, buffer, one * size);
	JS_Call(callback->ctx, callback->value, JS_NULL, 1, &value);
	JS_FreeValue(callback->ctx, value);
	return one * size;
}

size_t qr_builtin_curl_header(buffer, one, size, data)
	char *buffer;
	size_t one;
	size_t size;
	void *data;
{
	QrJSValue callback = data;
	JSValue value = JS_NewStringLen(callback->ctx, buffer, one * size);
	JS_Call(callback->ctx, callback->value, JS_NULL, 1, &value);
	JS_FreeValue(callback->ctx, value);
	return one * size;
}

JSValue qr_builtin_curl(ctx, this, argc, argv)
	JSContext *ctx;
	JSValue this;
	int argc;
	JSValue *argv;
{
	JSValue method_ = argv[0];
	
	JSValue scheme_ = JS_GetPropertyStr(ctx, argv[1], "protocol");
	JSValue host_ = JS_GetPropertyStr(ctx, argv[1], "hostname");
	JSValue port_ = JS_GetPropertyStr(ctx, argv[1], "port");
	JSValue path_ = JS_GetPropertyStr(ctx, argv[1], "pathname");
	JSValue query_ = JS_GetPropertyStr(ctx, argv[1], "search");
	
	JSValue headers_ = argv[4];
	
	JSValue callback_ = JS_DupValue(ctx, argv[3]);
	JSValue header_callback_ = JS_DupValue(ctx, argv[4]);
	JSValue done_callback_ = JS_DupValue(ctx, argv[5]);
	
	char *method = JS_ToCString(ctx, method_);
	char *scheme = JS_ToCString(ctx, scheme_);
	char *host = JS_ToCString(ctx, host_);
	char *port = JS_ToCString(ctx, port_);
	char *path = JS_ToCString(ctx, path_);
	char *query = JS_ToCString(ctx, query_);
	
	QrJSValue callback = malloc(sizeof *callback * 3);
	callback->ctx = ctx;
	callback->value = callback_;
	
	QrJSValue header_callback = callback + 1;
	header_callback->ctx = ctx;
	header_callback->value = header_callback_;
	
	QrJSValue done_callback = callback + 2;
	done_callback->ctx = ctx;
	done_callback->value = done_callback_;
	
	JSValue result = JS_UNDEFINED;
	
	CURL *curl = curl_easy_init();
	if (curl == NULL) result = JS_ThrowInternalError(ctx, "could not init cURL");
	CURLU *url = curl_url();
	if (url == NULL) result = JS_ThrowInternalError(ctx, "could not create cURL URL");
	
	QrDOM dom = JS_GetRuntimeOpaque(JS_GetRuntime(ctx));
	
	if (curl_multi_add_handle(qr_dom_curl(dom), curl) != CURLM_OK) result = JS_ThrowInternalError(ctx, "could not associate cURL handle");
	
	if (curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, &qr_builtin_curl_write) != CURLUE_OK) result = JS_ThrowInternalError(ctx, "could not set cURL callback");
	if (curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *) callback) != CURLUE_OK) result = JS_ThrowInternalError(ctx, "could not set cURL calback data");
	
	if (curl_easy_setopt(curl, CURLOPT_HEADERFUNCTION, &qr_builtin_curl_header) != CURLUE_OK) result = JS_ThrowInternalError(ctx, "could not set cURL header callback");
	if (curl_easy_setopt(curl, CURLOPT_HEADERDATA, (void *) header_callback) != CURLUE_OK) result = JS_ThrowInternalError(ctx, "could not set cURL header calback data");
	
	if (curl_easy_setopt(curl, CURLOPT_PRIVATE, (void *) callback) != CURLUE_OK) result = JS_ThrowInternalError(ctx, "could not set cURL data");
	
	JSValue length_ = JS_GetPropertyStr(ctx, headers_, "length");
	uint32_t length;
	qr_assert(JS_ToUint32(ctx, &length, length_) == 0);
	JS_FreeValue(ctx, length_);
	
	struct curl_slist *headers = NULL;
	
	for (uint32_t i = 0 ; i < length ; i++)
	{
		JSValue line_ = JS_GetPropertyUint32(ctx, headers_, i);
		char *line = JS_ToCString(ctx, line_);
		qr_assert(line != NULL);
		JS_FreeValue(ctx, line_);
		headers = curl_slist_append(headers, line);
	}
	
	if (curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers) != CURLUE_OK) result = JS_ThrowInternalError(ctx, "could not set curl headers");
	
	if (JS_IsUndefined(result))
	{
		char actual_scheme[strlen(scheme)];
		memcpy(actual_scheme, scheme, sizeof actual_scheme);
		actual_scheme[sizeof actual_scheme - 1] = '\0';
		
		if (curl_url_set(url, CURLUPART_SCHEME, actual_scheme, 0) != CURLUE_OK) result = JS_ThrowInternalError(ctx, "could not set URL scheme");
		if (curl_url_set(url, CURLUPART_HOST, host, 0) != CURLUE_OK) result = JS_ThrowInternalError(ctx, "could not set URL host");
		if (port[0] != '\0' && curl_url_set(url, CURLUPART_PORT, port, 0) != CURLUE_OK) result = JS_ThrowInternalError(ctx, "could not set URL port");
		if (curl_url_set(url, CURLUPART_PATH, path, 0) != CURLUE_OK) result = JS_ThrowInternalError(ctx, "could not set URL path");
		if (query[0] != '\0' && curl_url_set(url, CURLUPART_QUERY, query + 1, 0) != CURLUE_OK) result = JS_ThrowInternalError(ctx, "could not set URL query");
		
		if (curl_easy_setopt(curl, CURLOPT_CURLU, url) != CURLUE_OK) result = JS_ThrowInternalError(ctx, "could not set cURL URL");
	}
	else
	{
		curl_easy_cleanup(curl);
		curl_url_cleanup(url);
	}
	
	JS_FreeValue(ctx, host_);
	JS_FreeValue(ctx, scheme_);
	JS_FreeValue(ctx, port_);
	JS_FreeValue(ctx, path_);
	JS_FreeValue(ctx, query_);
	
	JS_FreeCString(ctx, method);
	JS_FreeCString(ctx, scheme);
	JS_FreeCString(ctx, host);
	JS_FreeCString(ctx, port);
	JS_FreeCString(ctx, path);
	JS_FreeCString(ctx, query);
	
	return result;
}

JSValue qr_builtin_url(ctx, this, argc, argv)
	JSContext *ctx;
	JSValue this;
	int argc;
	JSValue *argv;
{
	QrDOM dom = JS_GetRuntimeOpaque(JS_GetRuntime(ctx));
	return qr_dom_url(dom);
}

static JSCFunctionListEntry qr_builtin_array[] =
{
    JS_CFUNC_DEF("print", 0, &qr_builtin_print),
    
    JS_CFUNC_DEF("encode", 1, &qr_builtin_encode),
    JS_CFUNC_DEF("decode", 1, &qr_builtin_decode),
    
    JS_CFUNC_DEF("new_context", 0, &qr_builtin_new_context),
    JS_CFUNC_DEF("run", 2, &qr_builtin_run_script),
    
    JS_CFUNC_DEF("now", 0, &qr_builtin_now),
    
    JS_CFUNC_DEF("cURL", 6, &qr_builtin_curl),
    
    JS_CFUNC_DEF("url", 0, &qr_builtin_url),
};

JSCFunctionListEntry *qr_builtin = qr_builtin_array;
unsigned int qr_builtin_count = sizeof qr_builtin_array / sizeof *qr_builtin_array;
