// this file is part of the Quirk browser
// copyright 2021 by zamfofex
// license: GPLv3 or later

#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "builtin.h"

#include "dom.h"
#include "../util/syntax.h"

typedef struct QrSystemModuleEntry QrSystemModuleEntry;
struct QrSystemModuleEntry
{
	char *prefix;
	char *name;
	char *url;
	char *js;
	JSModuleDef *module;
};

static void qr_uncaught(JSContext *ctx);
static char *qr_normalize_system_module_name(JSContext *ctx, char *base, char *name, void *data);
static JSModuleDef *qr_load_system_module(JSContext *ctx, char *url, void *data);
static int qr_init_builtin(JSContext *ctx, JSModuleDef *module);
static QrSystemModuleEntry qr_builtin_entry(JSRuntime *rt);
static void qr_report_unhandled_rejection(JSContext *ctx, JSValue promise, JSValue reason, JS_BOOL is_handled, void *data);

void qr_report(text)
	char *text;
{
	size_t length = strlen(text);
	while (length > 0)
	{
		ssize_t count = write(2, text, length);
		if (count < 0) return;
		length -= count;
		text += count;
	}
}

char *qr_normalize_system_module_name(ctx, base, name, data)
	JSContext *ctx;
	char *base;
	char *name;
	void *data;
{
	QrSystemModuleEntry *modules = data;
	
	if (*name != '.')
	{
		for (size_t i = 0 ; modules[i].name != NULL ; i++)
		{
			if (strcmp(name, modules[i].name) == 0)
			if (strncmp(base, modules[i].prefix, strlen(modules[i].prefix)) == 0)
			{
				size_t length = strlen(modules[i].url);
				char *result = js_malloc(ctx, length + 1);
				qr_assert(result != NULL);
				memcpy(result, modules[i].url, length + 1);
				return result;
			}
		}
		qr_report("could not resolve system module '");
		qr_report(name);
		qr_report("'\n");
		qr_assert(0);
	}
	
	char *path = base;
	
	char *slash = strrchr(path, '/');
	size_t i = slash - path + 1;
	if (slash == NULL) i = 0;
	
	char *result = js_malloc(ctx, i + strlen(name) + 1);
	qr_assert(result != NULL);
	strncpy(result, path, i);
	
	for (;;)
	{
		if (name[0] == '/')
		{
			name++;
			continue;
		}
		
		if (name[0] == '.')
		if (name[1] == '/')
		{
			name += 2;
			continue;
		}
		
		if (name[0] == '.')
		if (name[1] == '.')
		if (name[2] == '/')
		{
			if (i != 0)
				i--;
			while (i != 0 && result[i - 1] != '/')
				i--;
			
			name += 3;
			continue;
		}
		
		for (;;)
		{
			result[i++] = *name;
			if (*name == '/') break;
			if (*name == '\0')
				return result;
			name++;
		}
	}
}

JSModuleDef *qr_load_system_module(ctx, url, data)
	JSContext *ctx;
	char *url;
	void *data;
{
	QrSystemModuleEntry *modules = data;
	
	for (size_t i = 0 ; modules[i].url != NULL ; i++)
	{
		if (strcmp(url, modules[i].url) == 0)
		{
			if (modules[i].module == NULL)
			{
				char *url = modules[i].url;
				char *js = modules[i].js;
				JSValue value = JS_Eval(ctx, js, strlen(js), url, JS_EVAL_TYPE_MODULE|JS_EVAL_FLAG_COMPILE_ONLY);
				qr_assert(!JS_IsException(value));
				modules[i].module = JS_VALUE_GET_PTR(value);
				
				JS_FreeValue(ctx, value);
			}
			return modules[i].module;
		}
	}
	
	qr_report("could not load system module '");
	qr_report(url);
	qr_report("'\n");
	qr_assert(0);
}

struct QrDOM
{
	JSRuntime *rt;
	JSContext *ctx;
	JSValue root;
	JSValue resume;
	JSValue exports;
	
	JSValue url;
	
	QrSystemModuleEntry *modules;
	
	CURLM *curl;
};

int qr_init_builtin(ctx, module)
	JSContext *ctx;
	JSModuleDef *module;
{
	return JS_SetModuleExportList(ctx, module, qr_builtin, qr_builtin_count);
}

QrSystemModuleEntry qr_builtin_entry(rt)
	JSRuntime *rt;
{
	JSContext *ctx = JS_NewContext(rt);
	qr_assert(ctx != NULL);
	
	JSModuleDef *builtin = JS_NewCModule(ctx, "builtin", &qr_init_builtin);
	qr_assert(builtin != NULL);
	qr_assert(JS_AddModuleExportList(ctx, builtin, qr_builtin, qr_builtin_count) == 0);
	
	QrSystemModuleEntry entry;
	entry.prefix = "quirk/";
	entry.name = "builtin";
	entry.url = "quirk/builtin";
	entry.module = builtin;
	
	return entry;
}

void qr_uncaught(ctx)
	JSContext *ctx;
{
	// todo
}

void qr_report_unhandled_rejection(ctx, promise, reason, is_handled, data)
	JSContext *ctx;
	JSValue promise;
	JSValue reason;
	JS_BOOL is_handled;
	void *data;
{
	// todo
}

QrDOM qr_dom(url)
	char *url;
{
	JSRuntime *rt = JS_NewRuntime();
	qr_assert(rt != NULL);
	
	QrDOM result = malloc(sizeof *result);
	qr_assert(result != NULL);
	
	JS_SetRuntimeOpaque(rt, result);
	result->rt = rt;
	
	result->curl = curl_multi_init();
	qr_assert(result->curl != NULL);
	
	QrSystemModuleEntry *modules = malloc(sizeof *modules);
	qr_assert(modules != NULL);
	
	modules[0] = qr_builtin_entry(rt);
	
	extern char _binary_js_modules_bin_start;
	extern char _binary_js_modules_bin_end;
	
	extern char _binary_dom_modules_bin_start;
	extern char _binary_dom_modules_bin_end;
	
	char *modules_start = &_binary_js_modules_bin_start;
	char *modules_end = &_binary_dom_modules_bin_end;
	
	for (size_t i = 1 ; modules_start != modules_end ; i++)
	{
		if (modules_start == &_binary_js_modules_bin_end)
			modules_start = &_binary_dom_modules_bin_start;
		
		char *prefix = modules_start;
		modules_start += strlen(prefix) + 1;
		
		char *name = modules_start;
		modules_start += strlen(name) + 1;
		
		char *url = modules_start;
		modules_start += strlen(url) + 1;
		
		modules = realloc(modules, (i + 2) * sizeof *modules);
		qr_assert(modules != NULL);
		
		modules[i].prefix = prefix;
		modules[i].name = name;
		modules[i].url = url;
		modules[i].js = modules_start;
		modules[i].module = NULL;
		
		modules[i + 1].name = NULL;
		
		modules_start += strlen(modules_start) + 1;
	}
	
	JS_SetModuleLoaderFunc(rt, &qr_normalize_system_module_name, &qr_load_system_module, modules);
	JS_SetHostPromiseRejectionTracker(rt, qr_report_unhandled_rejection, NULL);
	
	char *js = "import 'quirk/main.js'";
	
	JSContext *ctx = JS_NewContext(rt);
	qr_assert(ctx != NULL);
	
	result->ctx = ctx;
	result->url = JS_NewString(ctx, url);
	qr_assert(!JS_IsException(result->url));
	
	JSValue main = JS_Eval(ctx, js, strlen(js), "quirk/[start]", JS_EVAL_TYPE_MODULE);
	qr_assert(!JS_IsException(main));
	
	JSValue global = JS_GetGlobalObject(ctx);
	JSValue exports = JS_GetPropertyStr(ctx, global, "exports");
	qr_assert(!JS_IsException(exports));
	JS_FreeValue(ctx, global);
	
	JSValue resume = JS_GetPropertyStr(ctx, exports, "resume");
	qr_assert(!JS_IsException(resume));
	
	result->exports = exports;
	result->modules = modules;
	result->resume = resume;
	
	
	for (;;)
	{
		JSValue root = JS_GetPropertyStr(ctx, exports, "root");
		qr_assert(!JS_IsException(root));
		if (!JS_IsUndefined(root))
		{
			result->root = root;
			break;
		}
		
		qr_poll(result);
	}
	
	return result;
}

JSContext *qr_dom_ctx(dom)
	QrDOM dom;
{
	return dom->ctx;
}

JSValue qr_dom_root(dom)
	QrDOM dom;
{
	JSValue root = JS_Call(dom->ctx, dom->root, JS_NULL, 0, NULL);
	qr_assert(!JS_IsException(root));
	return root;
}

typedef struct QrJSValue *QrJSValue;
struct QrJSValue
{
	JSContext *ctx;
	JSValue value;
};

unsigned int qr_poll(dom)
	QrDOM dom;
{
	int handles = 0;
	curl_multi_perform(dom->curl, &handles);
	
	for (;;)
	{
		JSContext *other;
		int job = JS_ExecutePendingJob(dom->rt, &other);
		if (job == 0) break;
		qr_assert(job > 0);
	}
	
	JSValue time = JS_GetPropertyStr(dom->ctx, dom->exports, "time");
	int32_t ms;
	qr_assert(JS_ToInt32(dom->ctx, &ms, time) == 0);
	JS_FreeValue(dom->ctx, time);
	
	if (handles == 0 && ms == -1) return 0;
	
	if (ms == -1) ms = 120000;
	qr_assert(curl_multi_wait(dom->curl, NULL, 0, ms, NULL) == CURLE_OK);
	
	for (;;)
	{
		int count;
		CURLMsg *message = curl_multi_info_read(dom->curl, &count);
		if (message == NULL) break;
		
		if (message->msg != CURLMSG_DONE) continue;
		
		CURL *curl = message->easy_handle;
		qr_assert(curl_multi_remove_handle(dom->curl, curl) == CURLE_OK);
		
		QrJSValue callback;
		qr_assert(curl_easy_getinfo(curl, CURLINFO_PRIVATE, &callback) == CURLE_OK);
		JS_Call(callback->ctx, callback[2].value, JS_NULL, 0, NULL);
		JS_FreeValue(callback->ctx, callback[0].value);
		JS_FreeValue(callback->ctx, callback[1].value);
		JS_FreeValue(callback->ctx, callback[2].value);
		free(callback);
		
		curl_easy_cleanup(curl);
	}
	
	JS_Call(dom->ctx, dom->resume, JS_NULL, 0, NULL);
	
	return 1;
}

CURLM *qr_dom_curl(dom)
	QrDOM dom;
{
	return dom->curl;
}

JSValue qr_dom_url(dom)
	QrDOM dom;
{
	return dom->url;
}

void qr_dom_free(dom)
	QrDOM dom;
{
	qr_assert(curl_multi_cleanup(dom->curl) == CURLE_OK);
	JS_FreeValue(dom->ctx, dom->resume);
	JS_FreeValue(dom->ctx, dom->root);
	JS_FreeValue(dom->ctx, dom->url);
	JS_FreeContext(dom->ctx);
	JS_FreeRuntime(dom->rt);
}
