// this file is part of the Quirk browser
// copyright 2021 by zamfofex
// license: GPLv3 or later

#include "flow.h"

#include "../util/syntax.h"

#include <stdlib.h>

static void qr_flow_formal_intrinsic_sizes(QrBox box, QrAxis axis);
static void qr_flow_formal_sizes(QrBox box);
static void qr_flow_layout(QrBox box);

static void qr_runin_formal_intrinsic_sizes(QrBox box, QrAxis axis);
static void qr_runin_formal_sizes(QrBox box);
static void qr_runin_layout(QrBox box);

static void qr_word_formal_intrinsic_sizes(QrBox box, QrAxis axis);
static void qr_word_formal_sizes(QrBox box);
static void qr_word_layout(QrBox box);

static struct QrLayout qr_flow_struct = {&qr_flow_formal_intrinsic_sizes, &qr_flow_formal_sizes, &qr_flow_layout};
static struct QrLayout qr_runin_struct = {&qr_runin_formal_intrinsic_sizes, &qr_runin_formal_sizes, &qr_runin_layout};
static struct QrLayout qr_word_struct = {&qr_word_formal_intrinsic_sizes, &qr_word_formal_sizes, &qr_word_layout};

QrLayout qr_flow = &qr_flow_struct;
QrLayout qr_runin = &qr_runin_struct;
QrLayout qr_word = &qr_word_struct;

void qr_flow_formal_intrinsic_sizes(box, axis)
	QrBox box;
	QrAxis axis;
{
	QrNumber min;
	QrNumber max;
	
	if (axis == box->block_axis)
	{
		min = 0;
		max = qr_omega;
	}
	else
	{
		min = 0;
		max = 0;
		
		for (unsigned int i = 0 ; i < box->count ; i++)
		{
			QrBox child = box->children[i];
			
			QrSizes sizes = child->sizes[axis];
			
			if (sizes.actual_intrinsic.min > min)
				min = sizes.actual_intrinsic.min;
			if (sizes.actual_intrinsic.max < max)
				max = sizes.actual_intrinsic.max;
		}
	}
	
	QrNumber padding = 0;
	padding = qr_add(padding, box->paddings[axis].start);
	padding = qr_add(padding, box->paddings[axis].end);
	
	min = qr_add(min, padding);
	max = qr_add(max, padding);
	
	box->sizes[axis].formal_intrinsic.min = min;
	box->sizes[axis].formal_intrinsic.max = max;
}

void qr_flow_formal_sizes(parent)
	QrBox parent;
{
	for (unsigned int i = 0 ; i < parent->count ; i++)
	{
		QrBox box = parent->children[i];
		box->sizes[parent->block_axis].formal = box->sizes[parent->block_axis].given;
		
		if (box->sizes[parent->inline_axis].given != qr_omega)
		{
			box->sizes[parent->inline_axis].formal = box->sizes[parent->inline_axis].given;
		}
		else
		{
			QrNumber size = parent->sizes[parent->inline_axis].formal;
			
			size = qr_sub(size, parent->paddings[parent->inline_axis].start);
			size = qr_sub(size, parent->paddings[parent->inline_axis].end);
			
			box->sizes[parent->inline_axis].formal = size;
		}
	}
}

void qr_flow_layout(box)
	QrBox box;
{
	QrNumber offset = box->paddings[box->block_axis].start;
	QrNumber block_margin_end = 0;
	
	if (box->sizes[box->inline_axis].formal == qr_omega)
		box->sizes[box->inline_axis].actual = 0;
	else
		box->sizes[box->inline_axis].actual = box->sizes[box->inline_axis].formal;
	
	for (unsigned int i = 0 ; i < box->count ; i++)
	{
		QrBox child = box->children[i];
		
		QrNumber block_margin = child->margins[box->block_axis].start;
		
		if (i == 0 && box->paddings[box->block_axis].start == 0)
			block_margin = qr_sub(block_margin, box->margins[box->block_axis].start);
		
		if (block_margin < block_margin_end)
			block_margin = block_margin_end;
		
		block_margin_end = child->margins[box->block_axis].end;
		
		QrNumber inline_margin = child->margins[box->inline_axis].start;
		QrNumber inline_margin_end = child->margins[box->inline_axis].end;
		if (inline_margin == qr_omega)
		{
			inline_margin = box->sizes[box->inline_axis].actual - child->sizes[box->inline_axis].actual;
			if (inline_margin_end == qr_omega)
				inline_margin /= 2;
		}
		
		offset = qr_add(offset, block_margin);
		
		child->offsets[box->block_axis] = offset;
		child->offsets[box->inline_axis] = qr_add(box->paddings[box->inline_axis].start, inline_margin);
		
		offset = qr_add(offset, child->sizes[box->block_axis].actual);
	}
	
	if (box->sizes[box->block_axis].formal != qr_omega)
	{
		box->sizes[box->block_axis].actual = box->sizes[box->block_axis].formal;
	}
	else
	{
		if (box->paddings[box->block_axis].end == 0 && box->margins[box->block_axis].end < block_margin_end)
			offset = qr_add(offset, block_margin_end);
		offset = qr_add(offset, box->paddings[box->block_axis].end);
		box->sizes[box->block_axis].actual = offset;
	}
}

void qr_runin_formal_intrinsic_sizes(box, axis)
	QrBox box;
	QrAxis axis;
{
	QrNumber min;
	QrNumber max;
	
	if (axis == box->block_axis)
	{
		min = 0;
		max = qr_omega;
	}
	else
	{
		min = 0;
		max = box->gaps[axis] * box->count;
		
		for (unsigned int i = 0 ; i < box->count ; i++)
		{
			QrBox child = box->children[i];
			
			QrSizes sizes = child->sizes[axis];
			
			if (sizes.actual_intrinsic.min > min)
				min = sizes.actual_intrinsic.min;
			
			max = qr_add(max, sizes.actual_intrinsic.max);
		}
	}
	
	QrPadding paddings = box->paddings[axis];
	QrNumber padding = qr_add(paddings.start, paddings.end);
	
	min = qr_add(min, padding);
	max = qr_add(max, padding);
	
	box->sizes[axis].formal_intrinsic.min = min;
	box->sizes[axis].formal_intrinsic.max = max;
}

void qr_runin_formal_sizes(parent)
	QrBox parent;
{
	for (unsigned int i = 0 ; i < parent->count ; i++)
	{
		QrBox box = parent->children[i];
		box->sizes[qr_horizontal].formal = box->sizes[qr_horizontal].given;
		box->sizes[qr_vertical].formal = box->sizes[qr_vertical].given;
	}
}

void qr_runin_layout(box)
	QrBox box;
{
	box->sizes[box->inline_axis].actual = box->sizes[box->inline_axis].formal;
	
	QrNumber size = box->sizes[box->inline_axis].actual;
	
	QrNumber block_offset = box->paddings[box->block_axis].start;
	QrNumber inline_offset = box->paddings[box->inline_axis].start;
	QrNumber max = 0;
	int start = 1;
	
	for (unsigned int i = 0 ; i < box->count ; i++)
	{
		QrBox child = box->children[i];
		child->offsets[box->block_axis] = block_offset;
		child->offsets[box->inline_axis] = inline_offset;
		inline_offset = qr_add(inline_offset, child->sizes[box->inline_axis].given);
		
		if (child->sizes[box->block_axis].actual > max)
			max += child->sizes[box->block_axis].actual;
		
		if (inline_offset > size && start == 0)
		{
			block_offset = qr_add(block_offset, max);
			inline_offset = box->paddings[box->inline_axis].start;
			max = 0;
			start = 1;
			i--;
		}
		else
		{
			start = 0;
			inline_offset = qr_add(inline_offset, box->gaps[box->inline_axis]);
		}
	}
	
	box->offsets[box->block_axis] = block_offset;
	box->offsets[box->inline_axis] = inline_offset;
	box->sizes[box->block_axis].actual = qr_add(block_offset, max);
}

void qr_word_formal_intrinsic_sizes(box, axis)
	QrBox box;
	QrAxis axis;
{
	box->sizes[axis].formal_intrinsic.min = 0;
	box->sizes[axis].formal_intrinsic.max = qr_omega;
}

void qr_word_formal_sizes(parent)
	QrBox parent;
{
}

void qr_word_layout(box)
	QrBox box;
{
	box->sizes[qr_horizontal].actual = box->sizes[qr_horizontal].formal;
	box->sizes[qr_vertical].actual = box->sizes[qr_vertical].formal;
}
