// this file is part of the Quirk browser
// copyright 2021 by zamfofex
// license: GPLv3 or later

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "../layout/flow.h"
#include "../dom/dom.h"
#include "../util/syntax.h"
#include "../display/display.h"

#include "main.h"

typedef struct QrJSValue *QrJSValue;
struct QrJSValue
{
	JSContext *ctx;
	JSValue value;
};

typedef struct QrJSCString *QrJSCString;
struct QrJSCString
{
	JSContext *ctx;
	char *string;
	QrNumber advance;
};

static QrNumber qr_calc_normalized_sizes(
	JSContext *ctx,
	JSValue value_,
	QrNumber result,
	QrNumber empty,
	QrNumber full);

static QrNumber qr_css_calc_percentages(
	QrBox box,
	char *name,
	QrNumber result,
	QrNumber empty,
	QrNumber full);

static void qr_css_paddings(QrBox box, QrAxis axis, QrNumber percent);

static void qr_css_flow_normalized_sizes(QrBox box, QrAxis axis);
static void qr_css_flow_normalized_sizes_again(QrBox box, QrAxis axis, QrNumber percent, QrNumber padding_percent);
static QrBox qr_css_flow_box(JSContext *ctx, JSValue box_, QrDisplay display);

static void qr_css_word_normalized_sizes(QrBox box, QrAxis axis);
static void qr_css_word_normalized_sizes_again(QrBox box, QrAxis axis, QrNumber percent, QrNumber padding_percent);
static QrBox qr_css_word_box(JSContext *ctx, JSValue box_, QrDisplay display);

static void qr_css_runin_normalized_sizes(QrBox box, QrAxis axis);
static void qr_css_runin_normalized_sizes_again(QrBox box, QrAxis axis, QrNumber percent, QrNumber padding_percent);
static QrBox qr_css_runin_box(JSContext *ctx, JSValue box_, QrDisplay display);

static void qr_css_root_normalized_sizes(QrBox box, QrAxis axis);
static void qr_css_root_normalized_sizes_again(QrBox box, QrAxis axis, QrNumber percent, QrNumber padding_percent);

static QrBox qr_css_box(JSContext *ctx, JSValue box_, QrDisplay display);
static QrBox qr_root_box(void);
static void qr_css_boxes(JSContext *ctx, JSValue parent_, QrBox parent, QrDisplay display);
static void qr_css_box_free(QrBox box);
static void qr_css_boxes_free(QrBox box);

static void qr_merge_runin(QrBox box, QrNumber advance);
static QrBox qr_css_empty_box(void);

static void qr_rectangles(
	QrRectangle **rects,
	unsigned int *rect_count,
	QrText **texts,
	unsigned int *text_count,
	QrBox box,
	QrNumber x,
	QrNumber y);

int main(int argc, char **argv);

QrNumber qr_calc_normalized_sizes(ctx, value_, result, empty, full)
	JSContext *ctx;
	JSValue value_;
	QrNumber result;
	QrNumber empty;
	QrNumber full;
{
	JSValue operator_ = JS_GetPropertyStr(ctx, value_, "operator");
	
	if (!JS_IsUndefined(operator_))
	{
		// TODO
		JS_FreeValue(ctx, operator_);
	}
	else
	{
		JSValue unit_ = JS_GetPropertyStr(ctx, value_, "unit");
		JSValue number_ = JS_GetPropertyStr(ctx, value_, "value");
		
		int32_t number;
		qr_assert(JS_ToInt32(ctx, &number, number_) == 0);
		JS_FreeValue(ctx, number_);
		
		char *unit = JS_ToCString(ctx, unit_);
		qr_assert(unit != NULL);
		JS_FreeValue(ctx, unit_);
		
		if (strcmp(unit, "px") == 0)
			result = number;
		else if (strcmp(unit, "percent") == 0)
			result = number * full / 100 + empty;
		else if (strcmp(unit, "em") == 0)
			result = number * 16;
		
		JS_FreeCString(ctx, unit);
	}
	
	return result;
}

QrNumber qr_css_calc_percentages(box, name, result, empty, full)
	QrBox box;
	char *name;
	QrNumber result;
	QrNumber empty;
	QrNumber full;
{
	if (box->data == NULL) return 0;
	
	QrJSValue data = box->data;
	JSContext *ctx = data->ctx;
	JSValue properties_ = data->value;
	
	JSValue name_ = JS_NewString(ctx, name);
	JSValue get_ = JS_GetPropertyStr(ctx, properties_, "get");
	JSValue value_ = JS_Call(ctx, get_, properties_, 1, &name_);
	
	if (JS_IsNull(value_)) return result;
	
	JS_FreeValue(ctx, get_);
	JS_FreeValue(ctx, name_);
	
	char *value = JS_ToCString(ctx, value_);
	qr_assert(value != NULL);
	if (strcmp(value, "auto") == 0)
	{
		JS_FreeCString(ctx, value);
		return qr_omega;
	}
	JS_FreeCString(ctx, value);
	
	result = qr_calc_normalized_sizes(ctx, value_, result, empty, full);
	JS_FreeValue(ctx, value_);
	
	return result;
}

void qr_css_paddings(box, axis, percent)
	QrBox box;
	QrAxis axis;
	QrNumber percent;
{
	char *margin_start, *margin_end;
	char *padding_start, *padding_end;
	char *border_start, *border_end;
	
	if (axis == qr_horizontal)
	{
		margin_start = "margin-left";
		margin_end = "margin-right";
		padding_start = "padding-left";
		padding_end = "padding-right";
		border_start = "border-left-width";
		border_end = "border-right-width";
	}
	else
	{
		margin_start = "margin-top";
		margin_end = "margin-bottom";
		padding_start = "padding-top";
		padding_end = "padding-bottom";
		border_start = "border-top-width";
		border_end = "border-bottom-width";
	}
	
	box->margins[axis].start =
		qr_css_calc_percentages(box, margin_start, 0, 0, percent);
	box->margins[axis].end =
		qr_css_calc_percentages(box, margin_end, 0, 0, percent);
	
	box->paddings[axis].start =
		qr_css_calc_percentages(box, padding_start, 0, 0, percent) +
		qr_css_calc_percentages(box, border_start, 0, 0, percent);
	box->paddings[axis].end =
		qr_css_calc_percentages(box, padding_end, 0, 0, percent) +
		qr_css_calc_percentages(box, border_end, 0, 0, percent);
}

void qr_css_flow_normalized_sizes(box, axis)
	QrBox box;
	QrAxis axis;
{
	char *min, *max, *size;
	
	if (axis == qr_horizontal)
	{
		min = "min-width";
		max = "max-width";
		size = "width";
	}
	else
	{
		min = "min-height";
		max = "max-height";
		size = "height";
	}
	
	box->sizes[axis].given = qr_css_calc_percentages(box, size, qr_omega, qr_omega, qr_omega);
	box->sizes[axis].normalized.min = qr_css_calc_percentages(box, min, 0, 0, 0);
	box->sizes[axis].normalized.max = qr_css_calc_percentages(box, max, qr_omega, qr_omega, qr_omega);
}

void qr_css_flow_normalized_sizes_again(box, axis, percent, padding_percent)
	QrBox box;
	QrAxis axis;
	QrNumber percent;
	QrNumber padding_percent;
{
	char *min, *max, *size;
	
	if (axis == qr_horizontal)
	{
		min = "min-width";
		max = "max-width";
		size = "width";
	}
	else
	{
		min = "min-height";
		max = "max-height";
		size = "height";
	}
	
	box->sizes[axis].given = qr_css_calc_percentages(box, size, qr_omega, 0, percent);
	box->sizes[axis].normalized.min = qr_css_calc_percentages(box, min, 0, 0, percent);
	box->sizes[axis].normalized.max = qr_css_calc_percentages(box, max, qr_omega, 0, percent);
	
	qr_css_paddings(box, axis, padding_percent);
}

QrBox qr_css_flow_box(ctx, box_, display)
	JSContext *ctx;
	JSValue box_;
	QrDisplay display;
{
	QrBox box = malloc(sizeof *box);
	qr_assert(box != NULL);
	qr_css_boxes(ctx, box_, box, display);
	
	JSValue styles_ = JS_GetPropertyStr(ctx, box_, "styles");
	
	box->block_axis = qr_vertical;
	box->inline_axis = qr_horizontal;
	box->layout = qr_flow;
	
	QrJSValue data = malloc(sizeof *data);
	qr_assert(data != NULL);
	
	data->ctx = ctx;
	data->value = styles_;
	
	box->gaps[qr_horizontal] = 0;
	box->gaps[qr_vertical] = 0;
	
	box->margins[qr_horizontal].start = 0;
	box->margins[qr_horizontal].end = 0;
	
	box->margins[qr_vertical].start = 0;
	box->margins[qr_vertical].end = 0;
	
	box->paddings[qr_horizontal].start = 0;
	box->paddings[qr_horizontal].end = 0;
	
	box->paddings[qr_vertical].start = 0;
	box->paddings[qr_vertical].end = 0;
	
	box->data = data;
	box->normalized_sizes = &qr_css_flow_normalized_sizes;
	box->normalized_sizes_again = &qr_css_flow_normalized_sizes_again;
	
	return box;
}

void qr_css_word_normalized_sizes(box, axis)
	QrBox box;
	QrAxis axis;
{
	QrJSCString data = box->data;
	if (axis == box->inline_axis)
		box->sizes[axis].given = data->advance;
	else
		box->sizes[axis].given = 16;
	box->sizes[axis].normalized.min = 0;
	box->sizes[axis].normalized.max = qr_omega;
}

void qr_css_word_normalized_sizes_again(box, axis, percent, padding_percent)
	QrBox box;
	QrAxis axis;
	QrNumber percent;
	QrNumber padding_percent;
{
	qr_css_word_normalized_sizes(box, axis);
}

QrBox qr_css_word_box(ctx, box_, display)
	JSContext *ctx;
	JSValue box_;
	QrDisplay display;
{
	QrBox box = qr_css_empty_box();
	
	QrJSCString data = malloc(sizeof *data);
	qr_assert(data != NULL);
	
	data->ctx = ctx;
	data->string = JS_ToCString(ctx, box_);
	qr_assert(data->string != NULL);
	
	data->advance = (*display->advance)(display, data->string);
	box->data = data;
	
	box->layout = qr_word;
	box->normalized_sizes = &qr_css_word_normalized_sizes;
	box->normalized_sizes_again = &qr_css_word_normalized_sizes_again;
	
	return box;
}

void qr_css_runin_normalized_sizes(box, axis)
	QrBox box;
	QrAxis axis;
{
	box->sizes[axis].given = qr_omega;
	box->sizes[axis].normalized.min = 0;
	box->sizes[axis].normalized.max = qr_omega;
}

void qr_css_runin_normalized_sizes_again(box, axis, percent, padding_percent)
	QrBox box;
	QrAxis axis;
	QrNumber percent;
	QrNumber padding_percent;
{
	qr_css_runin_normalized_sizes(box, axis);
	qr_css_paddings(box, axis, padding_percent);
}

QrBox qr_css_runin_box(ctx, box_, display)
	JSContext *ctx;
	JSValue box_;
	QrDisplay display;
{
	QrBox box = qr_css_empty_box();
	qr_css_boxes(ctx, box_, box, display);
	box->gaps[box->inline_axis] = (*display->advance)(display, " ");
	return box;
}

void qr_css_root_normalized_sizes(box, axis)
	QrBox box;
	QrAxis axis;
{
	QrDisplay display = box->data;
	QrDisplaySize size = (*display->size)(display);
	if (axis == qr_horizontal)
	{
		box->sizes[axis].given = size.width;
		box->sizes[axis].normalized.min = size.width;
		box->sizes[axis].normalized.max = size.width;
	}
	else
	{
		box->sizes[axis].given = size.height;
		box->sizes[axis].normalized.min = size.height;
		box->sizes[axis].normalized.max = size.height;
	}
}

void qr_css_root_normalized_sizes_again(box, axis, percent, padding_percent)
	QrBox box;
	QrAxis axis;
	QrNumber percent;
	QrNumber padding_percent;
{
	qr_css_root_normalized_sizes(box, axis);
}

QrBox qr_css_empty_box()
{
	QrBox box = malloc(sizeof *box);
	qr_assert(box != NULL);
	
	box->children = NULL;
	box->count = 0;
	box->block_axis = qr_vertical;
	box->inline_axis = qr_horizontal;
	box->layout = qr_runin;
	box->data = NULL;
	
	box->normalized_sizes = &qr_css_runin_normalized_sizes;
	box->normalized_sizes_again = &qr_css_runin_normalized_sizes_again;
	
	box->gaps[qr_horizontal] = 0;
	box->gaps[qr_vertical] = 0;
	
	box->margins[qr_horizontal].start = 0;
	box->margins[qr_horizontal].end = 0;
	
	box->margins[qr_vertical].start = 0;
	box->margins[qr_vertical].end = 0;
	
	box->paddings[qr_horizontal].start = 0;
	box->paddings[qr_horizontal].end = 0;
	
	box->paddings[qr_vertical].start = 0;
	box->paddings[qr_vertical].end = 0;
	
	return box;
}

void qr_css_boxes(ctx, parent_, parent, display)
	JSContext *ctx;
	JSValue parent_;
	QrBox parent;
	QrDisplay display;
{
	JSValue children_ = JS_GetPropertyStr(ctx, parent_, "children");
	JSValue count_ = JS_GetPropertyStr(ctx, children_, "length");
	
	uint32_t count;
	qr_assert(JS_ToUint32(ctx, &count, count_) == 0);
	JS_FreeValue(ctx, count_);
	
	parent->children = malloc(count * sizeof *parent->children);
	qr_assert(parent->children != NULL);
	
	unsigned int j = 0;
	for (unsigned int i = 0 ; i < count ; i++)
	{
		JSValue box_ = JS_GetPropertyUint32(ctx, children_, i);
		parent->children[j] = qr_css_box(ctx, box_, display);
		JS_FreeValue(ctx, box_);
		if (parent->children[j] != NULL) j++;
	}
	parent->count = j;
	
	JS_FreeValue(ctx, children_);
}

QrBox qr_css_box(ctx, box, display)
	JSContext *ctx;
	JSValue box;
	QrDisplay display;
{
	if (JS_IsString(box))
		return qr_css_word_box(ctx, box, display);
	
	JSValue styles_ = JS_GetPropertyStr(ctx, box, "styles");
	JSValue get_property = JS_GetPropertyStr(ctx, styles_, "get");
	JSValue name = JS_NewString(ctx, "display");
	
	JSValue layout_ = JS_Call(ctx, get_property, styles_, 1, &name);
	char *layout = JS_ToCString(ctx, layout_);
	qr_assert(layout != NULL);
	
	JS_FreeValue(ctx, layout_);
	JS_FreeValue(ctx, name);
	JS_FreeValue(ctx, get_property);
	JS_FreeValue(ctx, styles_);
	
	QrBox result = NULL;
	
	if (strcmp(layout, "block") == 0 || strcmp(layout, "list-item") == 0)
		result = qr_css_flow_box(ctx, box, display);
	else if (strcmp(layout, "none") != 0)
		result = qr_css_runin_box(ctx, box, display);
	
	JS_FreeCString(ctx, layout);
	
	return result;
}

void qr_css_box_free(box)
	QrBox box;
{
	if (box->layout == qr_word)
	{
		QrJSCString data = box->data;
		JS_FreeCString(data->ctx, data->string);
		free(data);
	}
	else if (box->data != NULL)
	{
		QrJSValue data = box->data;
		JS_FreeValue(data->ctx, data->value);
		free(data);
	}
	
	free(box);
}

QrBox qr_root_box()
{
	QrBox box = malloc(sizeof *box);
	
	box->block_axis = qr_vertical;
	box->inline_axis = qr_horizontal;
	box->layout = qr_flow;
	
	box->gaps[qr_horizontal] = 0;
	box->gaps[qr_vertical] = 0;
	
	box->margins[qr_horizontal].start = 0;
	box->margins[qr_horizontal].end = 0;
	
	box->margins[qr_vertical].start = 0;
	box->margins[qr_vertical].end = 0;
	
	box->paddings[qr_horizontal].start = 0;
	box->paddings[qr_horizontal].end = 0;
	
	box->paddings[qr_vertical].start = 0;
	box->paddings[qr_vertical].end = 0;
	
	box->data = NULL;
	box->normalized_sizes = &qr_css_root_normalized_sizes;
	box->normalized_sizes_again = &qr_css_root_normalized_sizes_again;
	
	return box;
}

void qr_css_boxes_free(box)
	QrBox box;
{
	for (unsigned int i = 0 ; i < box->count ; i++)
		qr_css_boxes_free(box->children[i]);
	qr_css_box_free(box);
}

void qr_rectangles(rects, rect_count, texts, text_count, box, x, y)
	QrRectangle **rects;
	unsigned int *rect_count;
	QrText **texts;
	unsigned int *text_count;
	QrBox box;
	QrNumber x;
	QrNumber y;
{
	QrNumber ox = box->offsets[qr_horizontal];
	QrNumber oy = box->offsets[qr_vertical];
	if (ox == qr_omega) ox = 0;
	if (oy == qr_omega) oy = 0;
	x += ox; y += oy;
	
	rects[0][*rect_count] = malloc(sizeof ***rects);
	qr_assert(rects[0][*rect_count] != NULL);
	
	rects[0][*rect_count]->x = x;
	rects[0][*rect_count]->x = y;
	rects[0][*rect_count]->width = box->sizes[qr_horizontal].actual;
	rects[0][*rect_count]->height = box->sizes[qr_vertical].actual;
	rect_count[0]++;
	*rects = realloc(*rects, (*rect_count + 1) * sizeof **rects);
	qr_assert(*rects != NULL);
	rects[0][*rect_count] = NULL;
	
	if (box->layout == qr_word)
	{
		QrJSCString value = box->data;
		texts[0][*text_count] = malloc(sizeof ***texts);
		qr_assert(texts[0][*text_count] != NULL);
		texts[0][*text_count]->string = value->string;
		texts[0][*text_count]->x = x;
		texts[0][*text_count]->y = y;
		text_count[0]++;
		*texts = realloc(*texts, (*text_count + 1) * sizeof **texts);
		qr_assert(*texts != NULL);
		texts[0][*text_count] = NULL;
	}
	
	for (unsigned int i = 0 ; i < box->count ; i++)
		qr_rectangles(rects, rect_count, texts, text_count, box->children[i], x, y);
}

void qr_merge_runin(box, advance)
	QrBox box;
	QrNumber advance;
{
	QrBox last = NULL;
	for (unsigned int i = 0 ; i < box->count ; i++)
	{
		QrBox child = box->children[i];
		qr_merge_runin(child, advance);
		
		if (child->layout == qr_word && box->layout != qr_runin)
		{
			QrBox word = child;
			child = qr_css_empty_box();
			child->gaps[child->inline_axis] = advance;
			child->count = 1;
			child->children = malloc(sizeof *child->children);
			child->children[0] = word;
			box->children[i] = child;
		}
		
		if (child->layout != qr_runin)
		{
			last = NULL;
			continue;
		}
		if (last == NULL)
		{
			last = child;
			continue;
		}
		
		unsigned int count = last->count;
		last->count += child->count;
		last->children = realloc(last->children, last->count * sizeof *last->children);
		for (unsigned int i = 0 ; i < child->count ; i++)
			last->children[count + i] = child->children[i];
		qr_css_box_free(child);
		box->count--;
		memmove(box->children + i, box->children + i + 1, (box->count - i) * sizeof *box->children);
		box->children = realloc(box->children, box->count * sizeof *box->children);
		i--;
	}
	
	if (box->layout != qr_runin) return;
	
	for (unsigned int i = 0 ; i < box->count ; i++)
	{
		QrBox child = box->children[i];
		if (child->layout != qr_runin) continue;
		box->count--;
		unsigned int count = box->count;
		box->count += child->count;
		if (child->count != 0)
			box->children = realloc(box->children, box->count * sizeof *box->children);
		memmove(box->children + i + child->count, box->children + i + 1, (count - i) * sizeof *box->children);
		if (child->count == 0)
			box->children = realloc(box->children, box->count * sizeof *box->children);
		for (unsigned int j = 0 ; j < child->count ; i++, j++)
			box->children[i] = child->children[j];
		qr_css_box_free(child);
	}
}

int main(argc, argv)
	int argc;
	char **argv;
{
	QrDisplay display = qr_console(0, 1);
	
	char *url;
	
	if (argc == 1) url = "https://cricket.piapiac.org/software/";
	else if (argc == 2) url = argv[1];
	else exit(1);
	
	QrDOM dom = qr_dom(url);
	JSContext *ctx = qr_dom_ctx(dom);
	
	QrRectangle *rects = NULL;
	QrText *texts = NULL;
	QrBox root_box = NULL;
	
	for (;;)
	{
		JSValue root = qr_dom_root(dom);
		QrBox box = qr_css_flow_box(ctx, root, display);
		JS_FreeValue(ctx, root);
		
		if (box != NULL)
		{
			if (root_box != NULL)
			{
				free(rects);
				free(texts);
				free(root_box);
			}
			
			rects = malloc(sizeof *rects);
			qr_assert(rects != NULL);
			rects[0] = NULL;
			
			texts = malloc(sizeof *texts);
			qr_assert(texts != NULL);
			texts[0] = NULL;
			
			root_box = qr_root_box();
			root_box->data = display;
			root_box->count = 1;
			root_box->children = &box;
			
			qr_merge_runin(root_box, (*display->advance)(display, " "));
			qr_layout(root_box);
			
			unsigned int rect_count = 0;
			unsigned int text_count = 0;
			
			qr_rectangles(&rects, &rect_count, &texts, &text_count, root_box, 0, 0);
			(*display->display)(display, rects, texts);
		}
		if (qr_poll(dom) == 0) for (;;) pause();
	}
}
